# -*- coding: utf-8 -*-
"""
Created on Fri Jul  7 09:17:12 2023
"""
import pandas as pd
from functools import partial
from tqdm import tqdm
import logging

from french_cities.utils import init_pynsee
from pynsee.localdata import get_area_list
from pynsee.localdata import get_ascending_area
from pynsee.localdata import get_area_projection

logger = logging.getLogger(__name__)


def _get_cities_year_full(year: int, look_for: set = None) -> pd.DataFrame:
    """
    Download desired vintage of french official geographic code for cities
    and municipal districts from INSEE API; the obtained DataFrame contains
    two columns : the actual code of the territory and it's parent (which
    should be the same for an actual city, or the parent city for a municipal
    district).

    Parameters
    ----------
    year : int
        Desired vintage
    look_for : set, optional
        List of codes we are trying to project in the desired vintage.
        The default is None (will try to reach every available code).

    Returns
    -------
    cities : pd.DataFrame

            CODE  NEW_CODE
    0      01001     01001
    1      01002     01002
    2      01004     01004
    3      01005     01005
    4      01006     01006
         ...       ...
    34985  75116     75056
    34986  75117     75056
    34987  75118     75056
    34988  75119     75056
    34989  75120     75056

    """
    cities = _get_cities_year(year)
    data = [
        cities.assign(PARENT=cities["CODE"]),
        _get_subareas_year("arrondissementsMunicipaux", year, look_for),
        _get_subareas_year("communesAssociees", year, look_for),
        _get_subareas_year("communesDeleguees", year, look_for),
    ]
    cities = pd.concat(data, ignore_index=True)
    ix = cities[cities.CODE.isin(look_for)].index
    cities = (
        cities.loc[ix]
        .drop_duplicates(keep="first")
        .reset_index(drop=True)
        .rename({"PARENT": "NEW_CODE"}, axis=1)
    ).copy()
    return cities


def _get_cities_year(year: int) -> pd.DataFrame:
    """
    Download desired vintage of french official geographic code for cities
    from INSEE API; municipal districts are excluded by this API.

    Parameters
    ----------
    year : int
        Desired vintage

    Returns
    -------
    cog : pd.DataFrame

        CODE
    0  01001
    1  01002
    2  01004
    3  01005
    4  01006

    """
    date = f"{year}-01-01"
    cog = get_area_list(area="communes", date=date)
    try:
        cog = cog.drop("DATE_DELETION", axis=1)
    except KeyError:
        pass
    cog = cog.drop(
        [
            "URI",
            "AREA_TYPE",
            "DETERMINER_TYPE",
            "TITLE",
            "TITLE_SHORT",
            "DATE_CREATION",
        ],
        axis=1,
    )
    return cog


def _get_parents_from_serie(
    type_: str, codes: list, year: int
) -> pd.DataFrame:
    """
    Get territories' parents codes using INSEE API. The output will be a
    DataFrame containing two columns (one for the actual territory's code, the
    other containing it's parent's code)

    Parameters
    ----------
    type_ : str
        case sensitive, area type, any of ('arrondissement',
        'arrondissementMunicipal', 'circonscriptionTerritoriale', 'commune',
        'communeAssociee', 'communeDeleguee', 'departement',
        'district')
    codes : list
        any iterable of territories' official codes of type type_
    year : int
        Desired vintage

    Returns
    -------
    parents : pd.DataFrame
           CODE PARENT
       0  13201  13055
       1  13202  13055
       2  13203  13055
       3  13204  13055
       4  13205  13055

    """
    date = f"{year}-01-01"
    parents = []
    func = partial(get_ascending_area, area=type_, date=date, type="commune")
    for code in tqdm(codes, desc="get parent from insee", leave=False):
        parents.append(
            {"CODE": code, "PARENT": func(code=code).loc[0, "code"]}
        )

    parents = pd.DataFrame(parents)
    return parents


def _get_subareas_year(
    type_: str, year: int, look_for: set = None
) -> pd.DataFrame:
    """
    Download desired vintage of french official geographic code for "subcities"
    (ie municipal districts, associated cities, delegated cities from INSEE API
    and create a link-table allowing to retrieve the parent city's code)

    Parameters
    ----------
    type_ : str
        Type of "subcity", among "arrondissementsMunicipaux",
        'communesAssociees', 'communesDeleguees'
    year : int
        Desired vintage
    look_for : set, optional
        List of codes we are trying to project in the desired vintage.
        The default is None (will try to reach every available code).

    Returns
    -------
    subareas : pd.DataFrame

        CODE PARENT
    0  13201  13055
    1  13202  13055
    2  13203  13055
    3  13204  13055
    4  13205  13055

    """
    date = f"{year}-01-01"
    subareas = get_area_list(area=type_, date=date)
    try:
        subareas = subareas.drop("DATE_DELETION", axis=1)
    except KeyError:
        pass
    drop = [
        "URI",
        "AREA_TYPE",
        "DETERMINER_TYPE",
        "TITLE",
        "DATE_CREATION",
        "TITLE_SHORT",
    ]
    subareas = subareas.drop(drop, axis=1)

    if look_for:
        subareas = subareas[subareas.CODE.isin(look_for)]
    if subareas.empty:
        return pd.DataFrame()

    single_area = {
        "arrondissementsMunicipaux": "arrondissementMunicipal",
        "communesAssociees": "communeAssociee",
        "communesDeleguees": "communeDeleguee",
    }
    type_ = single_area[type_]
    parents = _get_parents_from_serie(type_, subareas.CODE.unique(), year)
    subareas = subareas.merge(parents, on="CODE", how="left")

    return subareas


def set_vintage(df: pd.DataFrame, year: int, field: str) -> pd.DataFrame:
    """
    Project (approximatively) the cities codes of a dataframe into a desired
    vintage.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing city codes
    year : int
        Year to project the dataframe's city codes into
    field : str
        Field (column) of dataframe containing the city code

    Returns
    -------
    pd.DataFrame
        Projected DataFrame

    """

    init_pynsee()

    renamed = False
    if field in {"CODE, NEW_CODE"}:
        renamed = True
        df = df.rename({field: f"temp_{field}"}, axis=1)
        field = f"temp_{field}"

    uniques = df[[field]].drop_duplicates(keep="first").dropna()

    if uniques.empty:
        return df

    cities = _get_cities_year_full(year, set(uniques[field]))

    # Uptodate cities (cities, municipal districts, delegated cities, ...)
    uniques = uniques.merge(
        cities, left_on=field, right_on="CODE", how="left"
    )
    uniques = uniques.rename({"NEW_CODE": "PROJECTED"}, axis=1)

    # Obsolete cities : merge, etc. : look for projection starting from an old
    # date, using INSEE API
    date = f"{year}-01-01"
    starting_dates = [
        "1943-01-01",
        "1960-01-01",
        "1980-01-01",
        "2000-01-01",
        "2010-01-01",
    ]

    def get_city(x):
        for date_init in starting_dates:
            # Hack to deactivate standard error log entries by pynsee which are
            # concieved with a valid date in mind.
            previous_level = logging.root.manager.disable
            logging.disable(logging.ERROR)
            df = get_area_projection(
                code=x, area="commune", date=date_init, dateProjection=date
            )
            try:
                df = df.drop("DATE_DELETION", axis=1)
            except (KeyError, AttributeError):
                pass
            logging.disable(previous_level)

            if df is not None:
                break
        try:
            return df.at[0, "code"]
        except Exception:
            logger.error(f"No projection found for city {x}")
            return None

    ix = uniques[uniques.PROJECTED.isnull()].index
    uniques.loc[ix, "PROJECTED"] = uniques.loc[ix, field].apply(get_city)

    uniques = dict(uniques[[field, "PROJECTED"]].values)

    df.loc[:, field] = df.loc[:, field].map(uniques)

    if renamed:
        df = df.rename({field: field.lstrip("temp_")}, axis=1)

    return df



# -*- coding: utf-8 -*-
"""
Created on Thu Jul  6 11:38:34 2023
"""
import pandas as pd
import io
from requests_cache import CachedSession
from requests import Session
from datetime import timedelta
import logging
import numpy as np
import time
from tqdm import tqdm
from pebble import ThreadPool
from pynsee.localdata import get_area_list
from rapidfuzz import fuzz, process
from unidecode import unidecode

from french_cities.utils import init_pynsee, patch_the_patch


logger = logging.getLogger(__name__)


def _process_departements_from_postal(
    df: pd.DataFrame, source: str, alias: str, session: Session = None
) -> pd.DataFrame:
    """
    Retrieve departement's code from postoffice code. Adds the result as a new
    column to dataframe under the label 'alias'. Uses the BAN (Base Adresse
    Nationale under the hood) and OpenDataSoft's freemium API in backoffice (
    mostly in case of "Cedex" codes)

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing postal codes
    source : str
        Field containing the postal codes
    alias : str
        Column to store the departements' codes unto
    session : Session, optional
        Web session. The default is None (and will use a CachedSession with
        30 days expiration)

    Returns
    -------
    df : pd.DataFrame
        Updated DataFrame with departement's codes

    """

    if not session:
        session = CachedSession(
            allowable_methods=("GET", "POST"),
            expire_after=timedelta(days=30),
        )

    postal_codes = df[[source]].drop_duplicates(keep="first")
    files = [
        (
            "data",
            (
                "data.csv",
                postal_codes.to_csv(index=False, encoding="utf8"),
                "text/csv; charset=utf-8",
            ),
        ),
        ("postcode", (None, source)),
        ("result_columns", (None, source)),
        ("result_columns", (None, "result_context")),
    ]

    with patch_the_patch():
        r = session.post(
            # recherche grâce à l'API de la BAN
            "https://api-adresse.data.gouv.fr/search/csv/",
            files=files,
        )

    if not r.ok:
        raise Exception(
            f"Failed to query BAN's API with {files=} - response was {r}"
        )
    result = pd.read_csv(io.BytesIO(r.content), dtype=str)
    result[alias] = (
        result["result_context"].str.split(",", expand=True)[0].str.strip(" ")
    )
    result = result.drop("result_context", axis=1)

    # where code is unknown, use Christian Quest Dataset with Cedex codes and
    # OpenDataSoft API (v2.1 contrairement à la doc disponible) en Freemium
    # https://www.data.gouv.fr/fr/datasets/liste-des-cedex/#_
    # https://public.opendatasoft.com/explore/dataset/correspondance-code-cedex-code-insee/information/?flg=fr&q=code%3D68013&lang=fr
    def get(x):
        while True:
            r = session.get(
                # recherche "en masse" grâce à l'API de la BAN
                "https://public.opendatasoft.com/api/explore/v2.1/catalog/datasets/correspondance-code-cedex-code-insee/records",
                params={
                    "select": "insee,libelle,nom_com",
                    "where": f"code={x}",
                    "limit": "10",
                    "offset": "0",
                    "timezone": "UTC",
                    "include_links": "false",
                    "include_app_metas": "false",
                },
            )
            if r.ok:
                break
            elif r.status_code == 429:
                time.sleep(1)
            else:
                logger.warning(
                    f"Error occured on code {x} on OpenDataSoft's API"
                )
                return None
        results = r.json()["results"]

        for dict_ in results:
            dict_.update({source: x})
        return results

    ix = result[result[alias].isnull()].index
    if len(ix) > 0:
        logger.info("postal codes unrecognized - maybe Cedex codes")
        args = result.loc[ix, source].dropna().tolist()
        results_cedex = []
        with tqdm(
            total=len(args), desc="Querying OpenDataSoft API", leave=False
        ) as pbar:
            with ThreadPool(10) as pool:
                future = pool.map(get, args)
                results_iterator = future.result()
                while True:
                    try:
                        this_result = next(results_iterator)
                        if this_result:
                            results_cedex += this_result
                    except StopIteration:
                        break
                    finally:
                        pbar.update(1)
        results_cedex = pd.DataFrame(results_cedex)

        # Select main city in case of discordant results. For instance, with
        # postcode=74105, you'll get:
        #    insee       nom_dep          libelle          nom_com
        # 0  74145  HAUTE-SAVOIE  ANNEMASSE CEDEX          Juvigny
        # 1  74012  HAUTE-SAVOIE  ANNEMASSE CEDEX        Annemasse
        # 2  74305  HAUTE-SAVOIE  ANNEMASSE CEDEX   Ville-la-Grand
        # 3  74298  HAUTE-SAVOIE  ANNEMASSE CEDEX  Vétraz-Monthoux

        if not results_cedex.empty:
            for f in ["libelle", "nom_com"]:
                results_cedex[f] = (
                    results_cedex[f]
                    .fillna("")
                    .str.upper()
                    .apply(unidecode)
                    .str.split(r"\W+")
                    .str.join(" ")
                    .str.strip(" ")
                    .replace("", np.nan)
                )
            results_cedex["libelle"] = results_cedex["libelle"].str.replace(
                r" CEDEX", ""
            )
            results_cedex["score"] = results_cedex[
                ["libelle", "nom_com"]
            ].apply(lambda xy: fuzz.token_set_ratio(*xy), axis=1)

            results_cedex = results_cedex.sort_values([source, "score"])
            results_cedex = results_cedex.drop_duplicates(source, keep="last")
            results_cedex = results_cedex.drop(
                ["nom_com", "score", "libelle"], axis=1
            )
            results_cedex = results_cedex.drop_duplicates()

            results_cedex = _process_departements_from_insee_code(
                results_cedex,
                source="insee",
                alias="dep_cedex",
                session=session,
            )
            result = result.merge(results_cedex, on=source, how="left")
            result.loc[ix, alias] = result.loc[ix, "dep_cedex"]
            result = result.drop(
                list(
                    set(results_cedex.columns)
                    - {
                        source,
                    }
                ),
                axis=1,
            )

    ix = result[result[alias].isnull()].index
    if len(ix) > 0:
        # Still no results -> assume we can use the first characters of
        # postcode anyway
        result.loc[ix, alias] = _process_departements_from_insee_code(
            result.loc[ix],
            source=source,
            alias="dep",
            session=session,
        )["dep"]

    logger.info("résultat obtenu")

    df = df.merge(result, on=source, how="left")

    return df


def _process_departements_from_insee_code(
    df: pd.DataFrame, source: str, alias: str, session: Session = None
) -> pd.DataFrame:
    """
    Compute departement's codes from official french cities codes (COG INSEE).
    Adds the result as a new column to dataframe under the label 'alias'.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing official cities codes
    source : str
        Field containing the official codes (INSEE COG)
    alias : str
        Column to store the departements' codes unto
    session : Session, optional
        Web session. The default is None (and will use a CachedSession with
        30 days expiration)
        **ignored argument, set only for coherence with
        _process_departements_from_postal**

    Returns
    -------
    df : pd.DataFrame
        Updated DataFrame with departement's codes

    """
    df[alias] = df[source].str[:2]

    ix = df[df[alias] == "97"].index
    df.loc[ix, alias] = df.loc[ix, source].str[:3]

    return df


def find_departements(
    df: pd.DataFrame,
    source: str,
    alias: str,
    type_code: str,
    session: Session = None,
) -> pd.DataFrame:
    """
    Compute departement's codes from postal or official codes (ie. INSEE COG)'
    Adds the result as a new column to dataframe under the label 'alias'.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing official cities codes
    source : str
        Field containing the post or official codes
    alias : str
        Column to store the departements' codes unto
    type_code : str
        Type of codes passed under `alias` label. Should be either 'insee' for
        official codes or 'postcode' for postal codes.
    session : Session, optional
        Web session. The default is None (and will use a CachedSession with
        30 days expiration)

    Raises
    ------
    ValueError
        If type_code not among "postcode", "insee".

    Returns
    -------
    df : pd.DataFrame
        Updated DataFrame with departement's codes

    """
    if type_code not in {"postcode", "insee"}:
        msg = (
            "type_code must be among ('postcode', 'insee') - "
            f"found {type_code} instead"
        )
        raise ValueError(msg)

    init_pynsee()

    df = df.copy()
    if type_code == "postcode":
        func = _process_departements_from_postal
    elif type_code == "insee":
        func = _process_departements_from_insee_code
    return func(df, source, alias, session)


def find_departements_from_names(
    df: pd.DataFrame, label: str, alias: str = "DEP_CODE"
) -> pd.DataFrame:
    """
    Retrieve departement's codes from their names.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing official departement's names
    label : str
        Field containing the label of the departements
    alias : str, optional
        Column to store the departements' codes unto.
        Default is "DEP_CODE"

    Returns
    -------
    df : pd.DataFrame
        Updated DataFrame with departement's codes

    """

    init_pynsee()
    candidates = get_area_list("departements")
    candidates = candidates[["CODE", "TITLE"]]
    candidates["TITLE"] = (
        candidates["TITLE"]
        .apply(unidecode)
        .str.upper()
        .str.replace(r"\W+", " ", regex=True)
    )
    candidates = dict(candidates[["TITLE", "CODE"]].values)
    candidates_keys = list(candidates.keys())

    df = df.copy()
    df["FORMATTED"] = (
        df[label]
        .apply(unidecode)
        .str.upper()
        .str.replace(r"\W+", " ", regex=True)
    )

    df[alias] = df["FORMATTED"].apply(
        lambda x: candidates[
            process.extractOne(
                x,
                candidates_keys,
                scorer=fuzz.ratio,
                score_cutoff=80,
            )[0]
        ]
    )
    df = df.drop("FORMATTED", axis=1)

    return df
